#!/usr/bin/env python3
import sys
import timeit
from pathlib import Path
from qualys_etl.etld_lib import etld_lib_functions
from qualys_etl.etld_lib import etld_lib_config
from qualys_etl.etld_lib import etld_lib_credentials

from qualys_etl.etld_knowledgebase import knowledgebase_03_extract
from qualys_etl.etld_knowledgebase import knowledgebase_05_transform_load_xml_to_sqlite
from qualys_etl.etld_knowledgebase import knowledgebase_06_distribution

global start_time
global stop_time


def knowledgebase_03_extract_wrapper(module_function=knowledgebase_03_extract, message=""):
    etld_lib_functions.logger.info(f"start {module_function} {message}")
    module_function.main()
    etld_lib_functions.logger.info(f"end   {module_function}")


def knowledgebase_05_transform_load_xml_to_sqlite_wrapper(
        module_function=knowledgebase_05_transform_load_xml_to_sqlite):
    etld_lib_functions.logger.info(f"start {module_function}")
    module_function.main()
    etld_lib_functions.logger.info(f"end   {module_function}")


def knowledgebase_06_distribution_wrapper(module_function=knowledgebase_06_distribution):
    etld_lib_functions.logger.info(f"start {module_function}")
    try:
        export_dir = etld_lib_config.kb_export_dir
        if Path(export_dir).is_dir():
            module_function.main()
        else:
            etld_lib_functions.logger.info(
                f"No distribution as directory does not exist: {export_dir}")
    except Exception as e:
        etld_lib_functions.logger.info(f"{module_function} ended with an {e}, "
                                       f"ignore if distribution is off.")
    etld_lib_functions.logger.info(f"end   {module_function}")


def knowledgebase_start_wrapper():
    global start_time
    start_time = timeit.default_timer()
    etld_lib_functions.logger.info(f"__start__ knowledgebase_etl_workflow {str(sys.argv)}")
    etld_lib_functions.logger.info(f"data directory: {etld_lib_config.qetl_user_data_dir}")
    etld_lib_functions.logger.info(f"config file:    {etld_lib_config.qetl_user_config_settings_yaml_file}")
    etld_lib_functions.logger.info(f"cred yaml file: {etld_lib_credentials.cred_file}")
    etld_lib_functions.logger.info(f"cookie file:    {etld_lib_credentials.cookie_file}")


def knowledgebase_end_wrapper():
    global start_time
    global stop_time

    stop_time = timeit.default_timer()
    etld_lib_functions.logger.info(f"runtime for knowledgebase_etl_workflow in seconds: {stop_time - start_time:,}")
    etld_lib_functions.logger.info(f"__end__ knowledgebase_etl_workflow {str(sys.argv)}")


def knowledgebase_etl_workflow():
    try:
        knowledgebase_start_wrapper()
        knowledgebase_03_extract_wrapper(message=f"kb_last_modified_after={etld_lib_config.kb_last_modified_after}")
        knowledgebase_05_transform_load_xml_to_sqlite_wrapper()
        knowledgebase_06_distribution_wrapper()
        knowledgebase_end_wrapper()
    except Exception as e:
        etld_lib_functions.logger.error(f"Error occurred, please investigate {sys.argv}")
        etld_lib_functions.logger.error(f"Exception: {e}")
        exit(1)


def main():
    knowledgebase_etl_workflow()


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name="knowledgebase_etl_workflow")
    etld_lib_config.main()
    etld_lib_credentials.main()
    main()



from qualys_etl.etld_lib import etld_lib_config as etld_lib_config
from qualys_etl.etld_lib import etld_lib_functions as etld_lib_functions
from qualys_etl.etld_archive import OLD_host_list_process_04_load_sqlite, OLD_knowledgebase_08_load_sqlite
from qualys_etl.etld_lib import etld_lib_sqlite_tables

global host_list_detection_updated_csv_columns


def host_list_detection_sqlite():
    global host_list_detection_updated_csv_columns
    sqlite_host_list_detection = etld_lib_sqlite_tables.SqliteObj(etld_lib_config.host_list_detection_sqlite_file)

    sqlite_host_list_detection.drop_and_recreate_table(etld_lib_config.host_list_detection_table_view_name,
                                                       host_list_detection_updated_csv_columns)

    sqlite_host_list_detection.bulk_insert_csv_file(etld_lib_config.host_list_detection_table_view_name,
                                                    etld_lib_config.host_list_detection_csv_file,
                                                    host_list_detection_updated_csv_columns,
                                                    "host list detection",
                                                    display_progress_counter_at_this_number=100000)
    sqlite_host_list_detection.commit_changes()
    sqlite_host_list_detection.close_connection()
    etld_lib_functions.logger.info(f"count rows added to {etld_lib_config.host_list_detection_table_view_name} table: "
                                   f"{sqlite_host_list_detection.count_rows_added_to_table:,}")


def start_msg_host_list_detection_sqlite():
    etld_lib_functions.logger.info(f"start")


def end_msg_host_list_detection_sqlite():
    etld_lib_functions.log_file_info(etld_lib_config.host_list_detection_csv_file, 'in')
    etld_lib_functions.log_file_info(etld_lib_config.host_list_detection_sqlite_file)
    etld_lib_functions.logger.info(f"end")


def combine_csv_columns_host_and_detection():
    global host_list_detection_updated_csv_columns
    csv_host_columns = etld_lib_config.host_list_detection_host_csv_columns()
    csv_host_detection_columns = etld_lib_config.host_list_detection_qids_csv_columns()
    csv_columns = csv_host_columns + csv_host_detection_columns
    idx = 0
    for column in csv_host_columns:
        if column in 'DETECTION_LIST':
            csv_columns.__delitem__(idx)
        idx = idx + 1
    host_list_detection_updated_csv_columns = csv_columns


def main():
    start_msg_host_list_detection_sqlite()
    combine_csv_columns_host_and_detection()
    host_list_detection_sqlite()
    OLD_knowledgebase_08_load_sqlite.kb_sqlite_file = etld_lib_config.host_list_detection_sqlite_file
    OLD_knowledgebase_08_load_sqlite.main()
    OLD_host_list_process_04_load_sqlite.host_list_sqlite_file = etld_lib_config.host_list_detection_sqlite_file
    OLD_host_list_process_04_load_sqlite.main()
    etld_lib_sqlite_tables.create_table_q_knowledgebase_in_host_list_detection(
        etld_lib_config.host_list_detection_sqlite_file)
    end_msg_host_list_detection_sqlite()


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name='host_list_detection_sqlite')
    etld_lib_config.main()
    main()



#!/usr/bin/env python3
from qualys_etl.etld_lib import etld_lib_config
from qualys_etl.etld_lib import etld_lib_spawn_etl
from qualys_etl.etld_asset_inventory import asset_inventory_02_workflow_manager


def main():
    etld_lib_config.set_path_qetl_user_home_dir()
    etld_lib_spawn_etl.log_dir = etld_lib_config.qetl_user_log_dir
    etld_lib_spawn_etl.log_file_path = etld_lib_config.asset_inventory_log_file
    etld_lib_spawn_etl.log_file_rotate_path = etld_lib_config.asset_inventory_log_rotate_file
    etld_lib_spawn_etl.lock_file = etld_lib_config.asset_inventory_lock_file
    etld_lib_spawn_etl.log_file_max_size = (1024 * 100000)          # 1024 * size = Max Meg Size
    etld_lib_spawn_etl.spawned_process_max_run_time = 604800        # seconds before terminate as there is an issue.
    etld_lib_spawn_etl.spawned_process_sleep_time = 5               # check every n seconds for spawned_process.is_alive()
    etld_lib_spawn_etl.spawned_process_count_to_status_update = 11  # print status spawned_process.is_alive() after n checks

    etld_lib_spawn_etl.etl_dir_to_monitor = etld_lib_config.qetl_user_data_dir
    etld_lib_spawn_etl.etl_files_max_time_between_changes = 1800

    etld_lib_spawn_etl.target_module_to_run = asset_inventory_02_workflow_manager.asset_inventory_etl_workflow
    etld_lib_spawn_etl.target_method_in_module_to_run = "asset_inventory_etl_workflow"
    etld_lib_spawn_etl.etl_main()


if __name__ == '__main__':
    main()



from qualys_etl.etld_lib import etld_lib_config as etld_lib_config
from qualys_etl.etld_lib import etld_lib_functions as etld_lib_functions
from qualys_etl.etld_lib import etld_lib_sqlite_tables as etld_lib_sqlite_tables

global host_list_sqlite_file


def host_list_sqlite():
    global host_list_sqlite_file
    sqlite_host_list = etld_lib_sqlite_tables.SqliteObj(host_list_sqlite_file)
    sqlite_host_list.drop_and_recreate_table(etld_lib_config.host_list_table_name,
                                             etld_lib_config.host_list_csv_columns(),
                                             key='ID')

    sqlite_host_list.bulk_insert_csv_file(etld_lib_config.host_list_table_name,
                                          etld_lib_config.host_list_csv_file,
                                          etld_lib_config.host_list_csv_columns(),
                                          "host list")

    sqlite_host_list.commit_changes()
    sqlite_host_list.close_connection()
    etld_lib_functions.logger.info(f"count rows added to {etld_lib_config.host_list_table_name} table: "
                                   f"{sqlite_host_list.count_rows_added_to_table:,}")


def start_msg_host_list_sqlite():
    etld_lib_functions.logger.info(f"start")


def end_msg_host_list_sqlite():
    etld_lib_functions.log_file_info(etld_lib_config.host_list_csv_file, 'input file')
    etld_lib_functions.log_file_info(host_list_sqlite_file)
    etld_lib_functions.logger.info(f"end")


def setup_vars():
    global host_list_sqlite_file
    try:
        host_list_sqlite_file
    except:
        host_list_sqlite_file = etld_lib_config.host_list_sqlite_file


def main():
    start_msg_host_list_sqlite()
    setup_vars()
    host_list_sqlite()
    end_msg_host_list_sqlite()


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name='host_list_sqlite')
    etld_lib_config.main()
    main()




import csv
import shelve
import json

from qualys_etl.etld_lib import etld_lib_config as etld_lib_config
from qualys_etl.etld_lib import etld_lib_functions as etld_lib_functions
global count_host_ids_written
global count_host_qids_written
global host_list_detection_item
global host_list_detection_csv_truncate_cell_limit
global host_list_detection_csv_file
global count_csv_records_written
global counter_obj


def prepare_host_info(csv_column, present_csv_cell_as_json=False):
    global host_list_detection_item
    # Prepare host_list_detection_item for csv row
    # Prepare Host List First, then at DETECTION_LIST
    if host_list_detection_item[csv_column] is None:
        host_list_detection_item[csv_column] = ""
    elif 'DATE' in csv_column:  # Prepare dates for use in Excel or Database
        host_list_detection_item[csv_column] = host_list_detection_item[csv_column].replace("T", " ").replace("Z", "")
    elif 'DETECTION_LIST' in csv_column:
        pass
    elif not isinstance(host_list_detection_item[csv_column], str):
        if present_csv_cell_as_json:
            host_list_detection_item[csv_column] = json.dumps(host_list_detection_item[csv_column])
        else:
            flatten = etld_lib_functions.flatten_nest(host_list_detection_item[csv_column])
            host_list_detection_item[csv_column] = ""
            for key in flatten.keys():
                host_list_detection_item[csv_column] = f"{host_list_detection_item[csv_column]}{key}:{flatten[key]}\n"


def prepare_csv_cell(csv_cell, csv_column, present_csv_cell_as_json=False):
    if csv_cell is None:
        csv_cell = ""
    elif 'DATE' in csv_column:  # Prepare dates for use in Excel or Database
        csv_cell = csv_cell.replace("T", " ").replace("Z", "")
    elif 'ASSET_GROUP_IDS' in csv_column:  # Prepare IDS for use in Excel
        if present_csv_cell_as_json:
            csv_cell = json.dumps(csv_cell)
        else:
            csv_cell = csv_cell.replace(",", "\n")
    elif not isinstance(csv_cell, str):  # Flatten Nested XML into String
        if present_csv_cell_as_json:
            csv_cell = json.dumps(csv_cell)
        else:
            flatten = etld_lib_functions.flatten_nest(csv_cell)
            csv_cell = ""
            for key in flatten.keys():
                csv_cell = f"{csv_cell}{key}:{flatten[key]}\n"
    return csv_cell


def write_csv_detections_for_one_host(csv_column, csv_headers, writer, csv_detection_columns):
    global host_list_detection_item
    global count_host_ids_written
    global count_host_qids_written
    global count_csv_records_written
    global counter_obj

    def write_csv_host_list_detection_row():
        csv_row = csv_headers.copy()
        for csv_header in csv_headers:
            if csv_header in host_list_detection_item.keys():
                csv_row[csv_header] = \
                    prepare_csv_cell(host_list_detection_item[csv_header], csv_header,
                                     etld_lib_config.host_list_detection_present_csv_cell_as_json)
                csv_row[csv_header], truncated_field_list = etld_lib_functions.truncate_csv_cell(
                    max_length=etld_lib_config.host_list_detection_csv_truncate_cell_limit,
                    csv_cell=host_list_detection_item[csv_header],
                    truncated_field_list="",
                    csv_column=csv_header)
            else:
                csv_row[csv_header] = ""
        writer.writerow(csv_row)

    if 'DETECTION_LIST' in csv_column:
        detection_list = host_list_detection_item[csv_column]['DETECTION']
        if isinstance(detection_list, dict):
            detection_list_for_one = [detection_list]
            detection_list = detection_list_for_one
        for detection in detection_list:
            for csv_detection_column in csv_detection_columns:
                if csv_detection_column in detection.keys():
                    if detection[csv_detection_column] is None:
                        host_list_detection_item[csv_detection_column] = ""
                    elif 'DATE' in csv_detection_column:  # Prepare dates for use in Excel or Database
                        host_list_detection_item[csv_detection_column] = \
                            detection[csv_detection_column].replace("T", " ").replace("Z", "")
                    else:
                        host_list_detection_item[csv_detection_column] = detection[csv_detection_column]
                else:
                    host_list_detection_item[csv_detection_column] = ""
            # Write Host + One Detection
            write_csv_host_list_detection_row()
            count_host_qids_written += 1
            count_csv_records_written += 1
            counter_obj.display_counter_to_log()


def host_list_detection_to_csv():
    global count_host_ids_written
    global count_host_qids_written
    global host_list_detection_csv_truncate_cell_limit
    global host_list_detection_csv_file
    global host_list_detection_item
    global count_csv_records_written
    global counter_obj

    host_list_detection_csv_file = etld_lib_config.host_list_detection_csv_file
    host_list_detection_csv_truncate_cell_limit = int(etld_lib_config.host_list_detection_csv_truncate_cell_limit)

    csv_host_columns = etld_lib_config.host_list_detection_host_csv_columns()
    csv_detection_columns = etld_lib_config.host_list_detection_qids_csv_columns()
    count_host_ids_written = 0
    count_host_qids_written = 0
    count_csv_records_written = 0
    counter_obj = \
        etld_lib_functions.DisplayCounterToLog(display_counter_at=10000,
                                               logger_func=etld_lib_functions.logger.info,
                                               display_counter_log_message=
                                               "count host list detection records written to csv")

    csv_headers = {}
    for header in csv_host_columns:
        csv_headers[header] = ""
    for header in csv_detection_columns:
        csv_headers[header] = ""
    csv_headers.__delitem__('DETECTION_LIST')
    csv_headers_list = csv_headers.keys()

    try:
        with open(host_list_detection_csv_file, 'w', newline='', encoding='utf-8') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=csv_headers_list, quoting=csv.QUOTE_ALL)
            writer.writeheader()
            with shelve.open(str(etld_lib_config.host_list_detection_shelve_file), flag='r') as shelve_database:
                for shelve_database_item in shelve_database:
                    host_list_detection_item = shelve_database[shelve_database_item]
                    count_host_ids_written += 1
                    for csv_column in csv_host_columns:
                        if csv_column in host_list_detection_item.keys():
                            prepare_host_info(csv_column, etld_lib_config.host_list_detection_present_csv_cell_as_json)
                            write_csv_detections_for_one_host(csv_column, csv_headers,
                                                              writer, csv_detection_columns)
    except Exception as e:
        etld_lib_functions.logger.error(f"Error writing to file: {str(host_list_detection_csv_file)}, "
                              f"please retry after fixing error")
        etld_lib_functions.logger.error(f"Exception: {e}")
        exit(1)


def start_msg_host_list_detection_csv():
    etld_lib_functions.logger.info(f"start")


def end_msg_host_list_detection_csv():
    global count_host_ids_written
    global count_host_qids_written
    global counter_obj
    etld_lib_functions.logger.info(f"count host written to csv: {count_host_ids_written:,}")
    counter_obj.display_final_counter_to_log()
    etld_lib_functions.log_file_info(etld_lib_config.host_list_detection_shelve_file, 'in')
    etld_lib_functions.log_dbm_info(etld_lib_config.host_list_detection_shelve_file)
    etld_lib_functions.log_file_info(etld_lib_config.host_list_detection_csv_file)
    etld_lib_functions.logger.info(f"end")


def host_list_detection_csv():
    global counter_obj
    host_list_detection_to_csv()


def main():
    start_msg_host_list_detection_csv()
    host_list_detection_csv()
    end_msg_host_list_detection_csv()


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name='host_list_detection_csv')
    etld_lib_config.main()
    main()



import sqlite3
import re
import csv
import sys
import json
from pathlib import Path
from qualys_etl.etld_lib import etld_lib_functions
from qualys_etl.etld_lib import etld_lib_config
global count_rows_added_to_table


def get_q_knowledgebase_min_max_dates(sqlite_file):
    min_date = None
    max_date = None
    try:
        conn = sqlite3.connect(sqlite_file)
        cursor = conn.cursor()
        select_statement = 'select min(LAST_SERVICE_MODIFICATION_DATETIME) MIN_DATE, ' \
                           'max(LAST_SERVICE_MODIFICATION_DATETIME) MAX_DATE ' \
                           'from Q_KnowledgeBase;'
        cursor.execute(select_statement)
        records = cursor.fetchall()
        for row in records:
            min_date = row[0]
            max_date = row[1]
        conn.close()
    except Exception as e:
        min_date = 'Not Found'
        max_date = 'Not Found'

    return min_date, max_date


def get_q_table_min_max_dates(sqlite_file, datetime_field_name, table_name):
    min_date = None
    max_date = None
    try:
        conn = sqlite3.connect(sqlite_file)
        cursor = conn.cursor()
        select_statement = f'select min({datetime_field_name}) MIN_DATE, ' \
                           f'max({datetime_field_name}) MAX_DATE ' \
                           f'from {table_name};'
        cursor.execute(select_statement)
        records = cursor.fetchall()
        for row in records:
            min_date = row[0]
            max_date = row[1]
        conn.close()
    except Exception as e:
        min_date = 'Not Found'
        max_date = 'Not Found'

    return min_date, max_date


def create_table_q_knowledgebase_in_host_list_detection():

    table_name = 'Q_KnowledgeBase_In_Host_List_Detection'
    kb_table_name = 'Q_KnowledgeBase'
    table_statement = f'CREATE TABLE {table_name} as ' \
                      f'select * from {kb_table_name} ' \
                      f'where QID in (select distinct(QID) from Q_Host_List_Detection);'
    try:

        conn = sqlite3.connect()
        conn.execute(f"DROP TABLE IF EXISTS {table_name};")
        conn.execute(table_statement)
        conn.execute(f"DROP TABLE IF EXISTS {kb_table_name};")
        conn.commit()
        conn.close()
        conn = sqlite3.connect(sqlite_file)
        conn.isolation_level = None
        conn.execute("VACUUM")
        conn.close()
    except Exception as e:
        etld_lib_functions.logger.error(f"Error creating table: {table_name}, please retry after fixing error")
        etld_lib_functions.logger.error(f"Exception: {e}")
        exit(1)


class SqliteObj:
    def __init__(self, sqlite_file, read_only=False):
        self.sqlite_file = str(sqlite_file)
        self.read_only = read_only
        self.connection = None
        self.cursor = None
        self.isolation_level = None
        self.count_rows_added_to_table = 0
        self.status_table_recreated = False
        self.open_connection_and_cursor(read_only=read_only)

    def open_connection_and_cursor(self, read_only=False, open_cursor=True, enable_row_factory=True):
        try:
            if read_only is True and Path(self.sqlite_file).exists:
                self.connection = sqlite3.connect(f'file:{str(self.sqlite_file)}?mode=ro', uri=True, timeout=300)
            else:
                self.connection = sqlite3.connect(database=self.sqlite_file, timeout=300)
            if enable_row_factory is True:
                self.connection.row_factory = sqlite3.Row
            if open_cursor is True:
                self.cursor = self.connection.cursor()

        except Exception as e:
            etld_lib_functions.logger.error(f"Error getting cursor obj: {self.sqlite_file}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)

    def attach_database_to_connection(self, database_as_name, database_sqlite_file: str, mode='ro'):
        attach_database_sql = f'ATTACH DATABASE ? AS {database_as_name}'
        # unsupported on redhat 8.4 self.connection.execute(attach_database_sql, (f'file:{str(database_sqlite_file)}?mode={mode}',))
        try:
            self.connection.execute(attach_database_sql, (str(database_sqlite_file),))
        except sqlite3.Error as err:
            etld_lib_functions.logger.error(f"sqlite3.Error attaching database: {self.sqlite_file}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {err}")
            exit(1)
        except Exception as e:
            etld_lib_functions.logger.error(f"Error attaching database: {self.sqlite_file}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)

    def get_table_columns(self, table_name) -> list:
        try:
            sql = f"select * from {table_name} where 1=0;"
            self.cursor.execute(sql)
            return [d[0] for d in self.cursor.description]
        except Exception as e:
            etld_lib_functions.logger.info(f"Table not found: {table_name}")
            etld_lib_functions.logger.info(f"Exception: {e}")
            return []

    def get_min_max_dates_from_table_column(self, table_name, column_name):
        select_statement = f'select min({column_name}) MIN_DATE, ' \
                           f'max({column_name}) MAX_DATE ' \
                           f'from {table_name};'
        self.cursor.execute(select_statement)
        records = self.cursor.fetchall()
        min_date=""
        max_date=""
        if len(records) == 1:
            min_date = records[0][0]
            max_date = records[0][1]
        return {'min_date': min_date, 'max_date': max_date}

    def execute_statement(self, statement):
        self.cursor.execute(f"{statement};")

    def reopen_connection_and_cursor_as_sqlite3_row_factory(self):
        self.close_connection()
        try:
            self.connection = sqlite3.connect(database=self.sqlite_file, timeout=300)
            self.connection.row_factory = sqlite3.Row
            self.cursor = self.connection.cursor()
        except Exception as e:
            etld_lib_functions.logger.error(f"Error getting row factory connnection cursor obj: {self.sqlite_file}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)

    def create_table(self,
                     table_name=None, csv_columns=[],
                     key=False, csv_column_types={}):
        try:
            if key is False:
                table_statement = self.create_table_statement_no_primary_key(table_name, csv_columns)
            else:
                table_statement = self.create_table_statement(key, table_name, csv_columns, csv_column_types)

            self.cursor.execute(table_statement)
            table_statement_one_line = table_statement.replace("\n", " ")
            etld_lib_functions.logger.info(f"sqlite create table statement: {table_statement_one_line}")
            self.connection.commit()
        except Exception as e:
            etld_lib_functions.logger.error(f"Error creating table: {table_name}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)

    def drop_and_recreate_table(self,
                                table_name=None, csv_columns=[],
                                key=False, csv_column_types={}):
        try:
            self.cursor.execute(f"DROP TABLE IF EXISTS {table_name};")
            if key is False:
                table_statement = \
                    self.create_table_statement_no_primary_key_and_column_types(
                        table_name, csv_columns, csv_column_types)
            else:
                table_statement = self.create_table_statement(key, table_name, csv_columns, csv_column_types)

            self.cursor.execute(table_statement)
            table_statement_one_line = table_statement.replace("\n", " ")
            etld_lib_functions.logger.info(f"sqlite create table statement: {table_statement_one_line}")
            self.connection.commit()
        except Exception as e:
            etld_lib_functions.logger.error(f"Error creating table: {table_name}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)

    @staticmethod
    def create_table_statement(primary_key, table_name, csv_columns, csv_column_types={}):
        table_statement = f"CREATE TABLE IF NOT EXISTS {table_name} (\n"
        primary_keys = ""
        if isinstance(primary_key, list):
            pass
        elif primary_key in csv_column_types.keys():
            table_statement = f"{table_statement}{primary_key} {csv_column_types[primary_key]} PRIMARY KEY NOT NULL,\n"
        else:
            table_statement = f"{table_statement}{primary_key} CHAR PRIMARY KEY NOT NULL,\n"

        for column in csv_columns:
            if isinstance(primary_key, list) or primary_key != column:
                if column in csv_column_types.keys():
                    column_type = csv_column_types[column]
                    table_statement = f"{table_statement}{column} {column_type},\n"
                else:
                    table_statement = f"{table_statement}{column} TEXT,\n"

        table_statement = f"{table_statement}Row_Last_Updated TEXT,\n"

        table_statement = re.sub(",\n$", "", table_statement)

        if isinstance(primary_key, list):
            primary_keys = ", ".join(primary_key)
            table_statement = f"{table_statement},\nPRIMARY KEY ({primary_keys})\n"

        table_statement = table_statement + ");"
        return table_statement

    @staticmethod
    def create_table_statement_no_primary_key_and_column_types(table_name, csv_columns, csv_column_types={}):
        table_statement = f"CREATE TABLE IF NOT EXISTS {table_name} (\n"
        for column in csv_columns:
            if column in csv_column_types.keys():
                column_type = csv_column_types[column]
                table_statement = f"{table_statement}{column} {column_type},\n"
            else:
                table_statement = f"{table_statement}{column} TEXT,\n"

        table_statement = f"{table_statement}Row_Last_Updated TEXT,\n"

        table_statement = re.sub(",\n$", "", table_statement)

        table_statement = table_statement + ");"
        return table_statement

    @staticmethod
    def create_table_statement_no_primary_key(table_name, csv_columns):
        table_statement_list = []
        table_statement_list.append(f"CREATE TABLE IF NOT EXISTS {table_name} (\n")
        for column in csv_columns:
            table_statement_list.append(f"{column} TEXT,\n")
        table_statement_list.append(f"Row_Last_Updated TEXT")
        table_statement_list.append(");")
        table_statement = "".join(table_statement_list)
        return table_statement

    def insert_or_replace_row(self, table_name: str, row: list):
        result = True
        try:
            self.cursor.execute(f"INSERT OR REPLACE INTO {table_name} "
                                f"VALUES({self.get_values_var(row)},datetime('now','utc'))",
                                row)
        except Exception as e:
            etld_lib_functions.logger.error(
                f"Error inserting or replacing row into table: {table_name}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)
        return result

    def insert_unique_row_ignore_duplicates(self, table_name: str, row: list):
        result = True
        try:
            self.cursor.execute(f"INSERT INTO {table_name} VALUES({self.get_values_var(row)},datetime('now','utc'))", row)
        except sqlite3.IntegrityError as err:
            if 'unique constraint failed' in str(err).lower():
                result = "duplicate"
                pass
            else:
                etld_lib_functions.logger.error(
                    f"Integrity Error inserting into: {table_name}, please retry after fixing error")
                etld_lib_functions.logger.error(f"Exception: {err}")
                exit(1)

        except sqlite3.Error as err:
            etld_lib_functions.logger.error(
                f"Error inserting row into table: {table_name}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {err}")
            exit(1)
        except Exception as e:
            etld_lib_functions.logger.error(
                f"Error inserting row into table: {table_name}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)
        return result

    def insert_unique_row(self, table_name: str, values_var: list, row: list, conflict: str):
        try:
            self.cursor.execute(f"INSERT INTO {table_name} VALUES({values_var},datetime('now','utc'))", row)
        except sqlite3.Error as err:
            etld_lib_functions.logger.error(
                f"Error inserting row into table: {table_name}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {err}")
            exit(1)
        except Exception as e:
            etld_lib_functions.logger.error(
                f"Error inserting row into table: {table_name}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)

    def commit_changes(self):
        try:
            self.connection.commit()
        except Exception as e:
            etld_lib_functions.logger.error(
                f"Error committing to: {self.sqlite_file}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)

    def close_connection(self):
        try:
            self.connection.close()
        except Exception as e:
            etld_lib_functions.logger.error(
                f"Error closing: {self.sqlite_file}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)

    def vacuum_database(self):
        try:
            self.isolation_level = self.connection.isolation_level
            self.connection.isolation_level = None
            self.connection.execute("VACUUM")
            self.connection.isolation_level = self.isolation_level
        except Exception as e:
            etld_lib_functions.logger.error(
                f"Error vacuuming database: {self.sqlite_file}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)

    def select_all_from_table(self, table_name):
        self.cursor.execute(f"SELECT * FROM {table_name}")
        return self.cursor.fetchall()

    def select_all_from_table_where_table_column_equals_value(
            self, table_name, table_column, where_table_column_equals_value):
        self.cursor.execute(f"SELECT * FROM {table_name} WHERE {table_column} LIKE '{where_table_column_equals_value}'")
        return self.cursor.fetchall()

    @staticmethod
    def get_values_var(csv_columns: list):
        values_var = ""
        for column in csv_columns:
            values_var = f"{values_var}?, "
        values_var = re.sub("\?, $", "?", values_var)
        return values_var

    @staticmethod
    def status_table_prepare_fields_for_update(status_name='HOST_LIST_LOAD_STATUS', batch_date="", batch_number=0,
                                               status='COMPLETED', total_rows_added=0):
        status_name = status_name
        status_detail_dict = {'BATCH_DATE': batch_date, 'BATCH_NUMBER': batch_number, 'STATUS': status,
                              'TOTAL_ROWS_ADDED': total_rows_added}
        status_count = int(batch_number)
        return status_name, status_detail_dict, status_count

    def status_table_insert_or_replace_row_in_sqlite(self,
                                                     table_name: str, status_name: str, status_detail_dict: dict,
                                                     status_count: int, commit_changes: bool = True):
        status_detail_json = json.dumps(status_detail_dict)
        row_to_insert = [status_name,
                         status_detail_json,
                         status_count]
        result = self.insert_or_replace_row(
            table_name=table_name,
            row=row_to_insert
        )
        if commit_changes is True:
            self.commit_changes()

    def status_table_select_status_row_from_sqlite(self,
                                                   table_name: str, table_column: str,
                                                   where_table_column_equals_value: str):
        result = self.select_all_from_table_where_table_column_equals_value(
            table_name=table_name,
            table_column=table_column,
            where_table_column_equals_value=where_table_column_equals_value
        )
        status_name = result[0]['STATUS_NAME']
        status_detail_dict = json.loads(result[0]['STATUS_DETAIL'])
        status_count = result[0]['STATUS_COUNT']
        return status_name, status_detail_dict, status_count

    def update_status_table(self,
                            batch_date: str,
                            batch_number: int,
                            total_rows_added_to_database: int,
                            status_table_name: str,
                            status_table_columns: list,
                            status_table_column_types: list,
                            status_name_column: str = 'HOST_LIST_LOAD_STATUS',
                            status_column: str = 'begin'):

        status_column_dict = {'begin': 'BEGIN LOADING BATCH',
                              'end': 'END LOADING BATCH',
                              'final': 'FINAL BATCH LOADED',
                              }

        if status_column in status_column_dict.keys():
            status_column_display = status_column_dict[status_column]
        else:
            status_column_display = status_column

        status_name, status_detail_dict, status_count = \
            self.status_table_prepare_fields_for_update(
                status_name=status_name_column,
                batch_date=batch_date,
                batch_number=batch_number,
                status=status_column_display,
                total_rows_added=total_rows_added_to_database)

        self.status_table_insert_or_replace_row_in_sqlite(
            table_name=status_table_name,
            status_name=status_name, status_detail_dict=status_detail_dict, status_count=status_count
        )
        status_name, status_detail_dict, status_count = \
            self.status_table_select_status_row_from_sqlite(
                table_name=status_table_name,
                table_column='STATUS_NAME',
                where_table_column_equals_value=status_name)
        etld_lib_functions.logger.info(f"Updated {status_table_name} - "
                                       f"status_name={status_name} - status_detail_dict={status_detail_dict}")

        return status_name, status_detail_dict, status_count

    def bulk_insert_csv_file(self, table_name, csv_file_name, csv_columns, message="",
                             display_progress_counter_at_this_number=10000):
        csv.field_size_limit(sys.maxsize)
        sqlite_rows_written = \
            etld_lib_functions.DisplayCounterToLog(
                display_counter_at=display_progress_counter_at_this_number,
                logger_func=etld_lib_functions.logger.info,
                display_counter_log_message=f"count {table_name} sqlite rows written")
        # etld_lib_functions.logger.info(f"csv.field_size_limit(sys.maxsize): {str(sys.maxsize)}")
        try:
            values_var = ""
            for column in csv_columns:
                values_var = f"{values_var}?, "
            values_var = re.sub("\?, $", "?", values_var)
            csv_headers = ""
            csv_row_list = []
            bulk_insert_count = 100000
            with open(csv_file_name, "r", encoding='utf-8') as csv_file:
                csv_reader = csv.reader(csv_file, delimiter=',')
                self.count_rows_added_to_table = 0
                csv_headers = next(csv_reader)  # Skip Header
                for row in csv_reader:
                    self.cursor.execute(
                        f"INSERT OR REPLACE INTO {table_name} VALUES({values_var},datetime('now','utc'))", row)
                    sqlite_rows_written.display_counter_to_log()
                    self.count_rows_added_to_table += 1
                self.commit_changes()
                sqlite_rows_written.display_final_counter_to_log()

        except sqlite3.Error as err:
            etld_lib_functions.logger.error(
                f"Error inserting row into table: {table_name}, please retry after fixing error")
            etld_lib_functions.logger.error(f"Exception: {err}")
            exit(1)
        except Exception as e:
            etld_lib_functions.logger.error(f"Error creating table from: {csv_file_name}, retry after fixing error")
            etld_lib_functions.logger.error(f"Row Number: {self.count_rows_added_to_table}")
            etld_lib_functions.logger.error(f"Exception: {e}")
            exit(1)



#!/usr/bin/env python3
import sys
import timeit
import time
from pathlib import Path
import multiprocessing
from qualys_etl.etld_lib import etld_lib_functions
from qualys_etl.etld_lib import etld_lib_config
from qualys_etl.etld_lib import etld_lib_credentials

from qualys_etl.etld_host_list_detection import host_list_detection_03_extract_controller
from qualys_etl.etld_host_list_detection import host_list_detection_05_transform_load_xml_to_sqlite
from qualys_etl.etld_host_list_detection import host_list_detection_06_distribution
from qualys_etl.etld_host_list import host_list_02_workflow_manager
from qualys_etl.etld_knowledgebase import knowledgebase_02_workflow_manager

global start_time
global stop_time


def host_list_02_workflow_manager_wrapper(module_function=host_list_02_workflow_manager, message=""):
    etld_lib_functions.logger.info(f"start {module_function} {message}")
    module_function.main()
    etld_lib_functions.logger.info(f"end   {module_function}")


def knowledgebase_02_workflow_manager_wrapper(
       module_function=knowledgebase_02_workflow_manager, message=""
):
    etld_lib_functions.logger.info(f"start {module_function} {message}")
    module_function.main()
    etld_lib_functions.logger.info(f"end   {module_function}")


def host_list_detection_03_extract_controller_wrapper(module_function=host_list_detection_03_extract_controller, message=""):
    etld_lib_functions.logger.info(f"start {module_function} {message}")
    module_function.main()
    etld_lib_functions.logger.info(f"end   {module_function}")


def host_list_detection_05_transform_load_xml_to_sqlite_wrapper(
        module_function=host_list_detection_05_transform_load_xml_to_sqlite):
    etld_lib_functions.logger.info(f"start {module_function}")
    module_function.main()
    etld_lib_functions.logger.info(f"end   {module_function}")


def host_list_detection_06_distribution_wrapper(module_function=host_list_detection_06_distribution):
    etld_lib_functions.logger.info(f"start {module_function}")
    module_function.main()
    etld_lib_functions.logger.info(f"end   {module_function}")


def host_list_detection_start_wrapper():
    global start_time
    start_time = timeit.default_timer()
    etld_lib_functions.logger.info(f"__start__ host_list_detection_etl_workflow {str(sys.argv)}")
    etld_lib_functions.logger.info(f"data directory: {etld_lib_config.qetl_user_data_dir}")
    etld_lib_functions.logger.info(f"config file:    {etld_lib_config.qetl_user_config_settings_yaml_file}")
    etld_lib_functions.logger.info(f"cred yaml file: {etld_lib_credentials.cred_file}")
    etld_lib_functions.logger.info(f"cookie file:    {etld_lib_credentials.cookie_file}")


def host_list_detection_end_wrapper():
    global start_time
    global stop_time

    stop_time = timeit.default_timer()
    etld_lib_functions.logger.info(f"runtime for host_list_detection_etl_workflow in seconds: {stop_time - start_time:,}")
    etld_lib_functions.logger.info(f"__end__ host_list_detection_etl_workflow {str(sys.argv)}")


def spawn_host_list_02_workflow_manager_wrapper():
    queue_of_batch_names_loaded_into_sqlite = multiprocessing.Queue()
    queue_process = \
        multiprocessing.Process(target=host_list_02_workflow_manager.main,
                                args=(queue_of_batch_names_loaded_into_sqlite,),
                                name='host_list_02_workflow_manager')
    return queue_process,queue_of_batch_names_loaded_into_sqlite


def spawn_knowledgebase_02_workflow_manager_wrapper():
    process = \
        multiprocessing.Process(target=knowledgebase_02_workflow_manager.main,
                                name='knowledgebase_02_workflow_manager')
    process.daemon = True
    process.start()
    return process


def host_list_detection_etl_workflow():
    host_list_detection_start_wrapper()

    try:
        knowledgebase_process = spawn_knowledgebase_02_workflow_manager_wrapper()
    except Exception as e:
        etld_lib_functions.logger.error(f"Error knowledgebase_02_workflow_manager_wrapper, please investigate {sys.argv}")
        etld_lib_functions.logger.error(f"Exception: {e}")
        exit(1)

    try:
        etld_lib_config.host_list_detection_vm_processed_after = etld_lib_config.host_list_vm_processed_after
        long_message = f"host_list_detection_vm_processed_after = " \
                       f"host_list_vm_processed_after={etld_lib_config.host_list_vm_processed_after}"
        host_list_02_workflow_manager_wrapper(message=long_message)
        while True:
            if knowledgebase_process.is_alive():
                time.sleep(1)
            else:
                if knowledgebase_process.exitcode != 0:
                    etld_lib_functions.logger.error(
                       "Error in knowledgebase processing.  Please investigate logs and repair issue.")
                    etld_lib_functions.logger.error(f"exit code for knowledgebase is: {knowledgebase_process.exitcode}")
                    raise Exception("knowledgebase_process exception")
                break

        # TODO create multiprocessing queue of host list batches loaded into sqlite.
        # TODO Use multiprocessing queue in host list detection to start processing host list as it is being downloaded
        # TODO instead of waiting.

    except Exception as e:
        etld_lib_functions.logger.error(f"Error host_list_02_workflow_manager_wrapper, please investigate {sys.argv}")
        etld_lib_functions.logger.error(f"Exception: {e}")
        exit(1)

    try:
        host_list_detection_03_extract_controller_wrapper(
            message=f"vm_processed_after={etld_lib_config.host_list_detection_vm_processed_after}")
    except Exception as e:
        etld_lib_functions.logger.error(f"Error host_list_detection_03_extract_controller_wrapper, please investigate {sys.argv}")
        etld_lib_functions.logger.error(f"Exception: {e}")
        exit(1)

    try:
        host_list_detection_06_distribution_wrapper()
    except Exception as e:
        etld_lib_functions.logger.error(f"Error host_list_detection_06_distribution_wrapper, please investigate {sys.argv}")
        etld_lib_functions.logger.error(f"Exception: {e}")
        exit(1)

    host_list_detection_end_wrapper()


def main():
    host_list_detection_etl_workflow()


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name='host_list_detection_etl_workflow')
    etld_lib_config.main()
    etld_lib_credentials.main()
    main()
